@@ -2,8 +2,13 @@ package client
 
 import (
 	"context"
+	"encoding/json"
 	"fmt"
+	"io/ioutil"
+	"log"
 	"net"
+	"os"
+	"os/exec"
 	"reflect"
 	"strings"
 	"time"
@@ -13,6 +18,7 @@ import (
 	rbacv1 "k8s.io/api/rbac/v1"
 	"k8s.io/apimachinery/pkg/api/errors"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
+	"k8s.io/apimachinery/pkg/util/intstr"
 
 	routev1 "github.com/openshift/api/route/v1"
 
@@ -71,23 +77,36 @@ func (cli *VanClient) RouterUpdateVersionInNamespace(ctx context.Context, hup bo
 		return false, err
 	}
 	site := config.GetSiteMetadata()
-	//compare to version of library running
+	// compare to version of library running
 	updateSite := false
 	if utils.LessRecentThanVersion(Version, site.Version) {
 		// site is newer than client library, cannot update
 		return false, fmt.Errorf("Site (%s) is newer than library (%s); cannot update", site.Version, Version)
 	}
-	rename := false
+	renameFor050 := false
+	addClaimsSupport := false
+	addMultiportServices := false
 	inprogress, originalVersion, err := cli.isUpdating(namespace)
 	if err != nil {
 		return false, err
 	}
 	if inprogress {
-		rename = utils.LessRecentThanVersion(originalVersion, "0.5.0")
+		renameFor050 = utils.LessRecentThanVersion(originalVersion, "0.5.0")
+		addClaimsSupport = utils.LessRecentThanVersion(originalVersion, "0.7.0")
+		addMultiportServices = utils.LessRecentThanVersion(originalVersion, "0.8.0")
+	} else {
+		originalVersion = site.Version
 	}
 	if utils.MoreRecentThanVersion(Version, site.Version) || (utils.EquivalentVersion(Version, site.Version) && Version != site.Version) {
-		if !inprogress && utils.LessRecentThanVersion(site.Version, "0.5.0") {
-			rename = true
+		if !inprogress {
+			if utils.LessRecentThanVersion(originalVersion, "0.7.0") {
+				addClaimsSupport = true
+				renameFor050 = utils.LessRecentThanVersion(originalVersion, "0.5.0")
+			}
+			if utils.LessRecentThanVersion(originalVersion, "0.8.0") {
+				addMultiportServices = true
+			}
+
 			err = cli.updateStarted(site.Version, namespace, configmap.ObjectMeta.OwnerReferences)
 			if err != nil {
 				return false, err
@@ -113,8 +132,8 @@ func (cli *VanClient) RouterUpdateVersionInNamespace(ctx context.Context, hup bo
 	usingRoutes := false
 	consoleUsesLoadbalancer := false
 	routerExposedAsIp := false
-	if rename {
-		//create new resources (as copies of old ones)
+	if renameFor050 {
+		// create new resources (as copies of old ones)
 		// services
 		_, err = kube.CopyService("skupper-messaging", types.LocalTransportServiceName, map[string]string{}, namespace, cli.KubeClient)
 		if err != nil && !errors.IsAlreadyExists(err) {
@@ -125,7 +144,7 @@ func (cli *VanClient) RouterUpdateVersionInNamespace(ctx context.Context, hup bo
 			return false, err
 		}
 		servingCertsAnnotation := map[string]string{
-			"service.alpha.openshift.io/serving-cert-secret-name": types.OauthConsoleSecret,
+			"service.alpha.openshift.io/serving-cert-secret-name": types.ConsoleServerSecret,
 		}
 		controllerSvc, err := kube.CopyService("skupper-controller", types.ControllerServiceName, servingCertsAnnotation, namespace, cli.KubeClient)
 		if err != nil && !errors.IsAlreadyExists(err) {
@@ -134,7 +153,7 @@ func (cli *VanClient) RouterUpdateVersionInNamespace(ctx context.Context, hup bo
 		if controllerSvc != nil {
 			consoleUsesLoadbalancer = controllerSvc.Spec.Type == corev1.ServiceTypeLoadBalancer
 		}
-		//update annotation on skupper-router-console if it exists
+		// update annotation on skupper-router-console if it exists
 		routerConsoleService, err := cli.KubeClient.CoreV1().Services(namespace).Get(types.RouterConsoleServiceName, metav1.GetOptions{})
 		if err == nil {
 			if routerConsoleService.ObjectMeta.Annotations == nil {
@@ -176,7 +195,7 @@ func (cli *VanClient) RouterUpdateVersionInNamespace(ctx context.Context, hup bo
 
 		usingRoutes, err = cli.usingRoutes(namespace)
 		if usingRoutes {
-			//no need to regenerate certificate as route names have not changed
+			// no need to regenerate certificate as route names have not changed
 			err = kube.CopySecret("skupper-internal", types.SiteServerSecret, namespace, cli.KubeClient)
 			if err != nil && !errors.IsAlreadyExists(err) {
 				return false, err
@@ -298,7 +317,7 @@ func (cli *VanClient) RouterUpdateVersionInNamespace(ctx context.Context, hup bo
 		}
 
 		if cli.RouteClient != nil {
-			//routes: skupper-controller -> skupper
+			// routes: skupper-controller -> skupper
 			original, err := cli.RouteClient.Routes(namespace).Get("skupper-controller", metav1.GetOptions{})
 			if err == nil {
 				route := &routev1.Route{
@@ -327,7 +346,7 @@ func (cli *VanClient) RouterUpdateVersionInNamespace(ctx context.Context, hup bo
 			} else if !errors.IsNotFound(err) {
 				return false, err
 			}
-			//need to update edge and inter-router routes to point at different service:
+			// need to update edge and inter-router routes to point at different service:
 			err = kube.UpdateTargetServiceForRoute(types.EdgeRouteName, types.TransportServiceName, namespace, cli.RouteClient)
 			if err != nil {
 				return false, err
@@ -344,8 +363,8 @@ func (cli *VanClient) RouterUpdateVersionInNamespace(ctx context.Context, hup bo
 		return false, err
 	}
 	updateRouter := false
-	if rename {
-		//update deployment
+	if renameFor050 {
+		// update deployment
 		// - serviceaccount
 		router.Spec.Template.Spec.ServiceAccountName = types.TransportServiceAccountName
 		// - mounted secrets:
@@ -364,7 +383,7 @@ func (cli *VanClient) RouterUpdateVersionInNamespace(ctx context.Context, hup bo
 	}
 	if updateRouter || updateSite || hup {
 		if !updateRouter {
-			//need to trigger a router redployment to pick up the revised metadata field
+			// need to trigger a router redployment to pick up the revised metadata field
 			touch(router)
 			updateRouter = true
 		}
@@ -382,17 +401,59 @@ func (cli *VanClient) RouterUpdateVersionInNamespace(ctx context.Context, hup bo
 		return false, err
 	}
 	updateController := false
-	if rename {
-		//update deployment
+	if renameFor050 {
+		// update deployment
 		// - serviceaccount
 		controller.Spec.Template.Spec.ServiceAccountName = types.ControllerServiceAccountName
 		// - mounted secrets:
 		kube.UpdateSecretVolume(&controller.Spec.Template.Spec, "skupper", types.LocalClientSecret)
-		kube.UpdateSecretVolume(&controller.Spec.Template.Spec, "skupper-controller-certs", types.OauthConsoleSecret)
+		kube.UpdateSecretVolume(&controller.Spec.Template.Spec, "skupper-controller-certs", types.ConsoleServerSecret)
 		// -oauth proxy sidecar
 		updateOauthProxyServiceAccount(&controller.Spec.Template.Spec, types.ControllerServiceAccountName)
 		updateController = true
 	}
+	if addClaimsSupport {
+		err = kube.UpdateRole(namespace, types.ControllerRoleName, types.ControllerPolicyRule, cli.KubeClient)
+		if err != nil {
+			return false, err
+		}
+		if !config.IsEdge() {
+			err = cli.addClaimsPortsToControllerService(ctx, namespace)
+			if err != nil {
+				return false, err
+			}
+			if usingRoutes {
+				err = cli.createClaimsRedemptionRoute(ctx, namespace)
+				if err != nil {
+					return false, err
+				}
+			}
+			var owner *metav1.OwnerReference
+			if len(controller.ObjectMeta.OwnerReferences) > 0 {
+				owner = &controller.ObjectMeta.OwnerReferences[0]
+			}
+			err = cli.createClaimsServerSecret(ctx, namespace, owner, usingRoutes)
+			if err != nil {
+				return false, err
+			}
+			kube.AppendSecretVolume(&controller.Spec.Template.Spec.Volumes, &controller.Spec.Template.Spec.Containers[0].VolumeMounts, types.ClaimsServerSecret, "/etc/service-controller/certs/")
+			updateController = true
+		}
+	}
+	if addMultiportServices {
+		// disabling the controller
+		controller, err = setAndWaitControllerReplicas(cli, 0)
+		if err != nil {
+			return false, err
+		}
+		if err = multiportConvertServices(ctx, cli); err != nil {
+			return false, err
+		}
+		if err = updateGatewayMultiport(ctx, cli); err != nil {
+			return false, err
+		}
+		updateController = true
+	}
 	desiredControllerImage := GetServiceControllerImageName()
 	if controller.Spec.Template.Spec.Containers[0].Image != desiredControllerImage {
 		controller.Spec.Template.Spec.Containers[0].Image = desiredControllerImage
@@ -400,10 +461,12 @@ func (cli *VanClient) RouterUpdateVersionInNamespace(ctx context.Context, hup bo
 	}
 	if updateController || hup {
 		if !updateController {
-			//trigger redeployment of service-controller to pick up latest image
+			// trigger redeployment of service-controller to pick up latest image
 			touch(controller)
 			updateController = true
 		}
+		replicas := int32(1)
+		controller.Spec.Replicas = &replicas
 		_, err = cli.KubeClient.AppsV1().Deployments(namespace).Update(controller)
 		if err != nil {
 			return false, err
@@ -426,8 +489,8 @@ func (cli *VanClient) RouterUpdateVersionInNamespace(ctx context.Context, hup bo
 			}
 		}
 	}
-	if rename {
-		//delete old resources
+	if renameFor050 {
+		// delete old resources
 		if cli.RouteClient != nil {
 			err = cli.RouteClient.Routes(namespace).Delete("skupper-controller", &metav1.DeleteOptions{})
 			if err != nil && !errors.IsNotFound(err) {
@@ -440,9 +503,9 @@ func (cli *VanClient) RouterUpdateVersionInNamespace(ctx context.Context, hup bo
 			"skupper-controller",
 		}
 		if usingRoutes {
-			//only delete skupper-internal if using
-			//routes, as otherwise previously issued
-			//tokens will reference it
+			// only delete skupper-internal if using
+			// routes, as otherwise previously issued
+			// tokens will reference it
 			services = append(services, "skupper-internal")
 		}
 		for _, service := range services {
@@ -505,6 +568,206 @@ func (cli *VanClient) RouterUpdateVersionInNamespace(ctx context.Context, hup bo
 	}
 	return updateRouter || updateController || updateSite, nil
 }
+func setAndWaitControllerReplicas(cli *VanClient, replicas int32) (*appsv1.Deployment, error) {
+	controller, err := cli.KubeClient.AppsV1().Deployments(cli.Namespace).Get(types.ControllerDeploymentName, metav1.GetOptions{})
+	if *controller.Spec.Replicas > 0 {
+		controller.Spec.Replicas = &replicas
+		_, err = cli.KubeClient.AppsV1().Deployments(cli.Namespace).Update(controller)
+		controller, err = kube.WaitDeploymentReadyReplicas(types.ControllerDeploymentName, cli.Namespace, int(replicas), cli.KubeClient, time.Minute, time.Second)
+		if err != nil {
+			return controller, err
+		}
+	}
+	return controller, err
+}
+
+func multiportConvertServices(ctx context.Context, cli *VanClient) error {
+	servicesCm, err := cli.KubeClient.CoreV1().ConfigMaps(cli.Namespace).Get(types.ServiceInterfaceConfigMap, metav1.GetOptions{})
+	if err != nil {
+		return err
+	}
+	v1Svcs := []types.ServiceInterfaceV1{}
+	for _, v := range servicesCm.Data {
+		v1Svc := types.ServiceInterfaceV1{}
+		err = json.Unmarshal([]byte(v), &v1Svc)
+		if err != nil {
+			return err
+		}
+		v1Svcs = append(v1Svcs, v1Svc)
+	}
+	outBytes, _ := json.Marshal(v1Svcs)
+	defs := &types.ServiceInterfaceList{}
+	err = defs.ConvertFrom(string(outBytes))
+	if err != nil {
+		return err
+	}
+	for _, svc := range *defs {
+		svcBytes, _ := json.Marshal(svc)
+		servicesCm.Data[svc.Address] = string(svcBytes)
+		_, err = cli.KubeClient.CoreV1().ConfigMaps(cli.Namespace).Update(servicesCm)
+		if err != nil {
+			return err
+		}
+		servicesCm, _ = cli.KubeClient.CoreV1().ConfigMaps(cli.Namespace).Get(types.ServiceInterfaceConfigMap, metav1.GetOptions{})
+	}
+
+	return err
+}
+
+func updateGatewayMultiport(ctx context.Context, cli *VanClient) error {
+	// retrieving all service definitions
+	svcList, _ := cli.ServiceInterfaceList(ctx)
+	svcMap := map[string]*types.ServiceInterface{}
+	for _, svc := range svcList {
+		svcMap[svc.Address] = svc
+	}
+	gwList, _ := cli.GatewayList(ctx)
+	for _, gw := range gwList {
+		// updating local gateways
+		gatewayDir := getDataHome() + "/skupper/" + gw.GatewayName
+		newGatewayDir := getDataHome() + gatewayClusterDir + gw.GatewayName
+		// create the new base dir for gateways (and ignore errors if it already exists)
+		_ = os.MkdirAll(getDataHome()+gatewayClusterDir, 0755)
+		gd, err := os.Stat(gatewayDir)
+		ngd, nerr := os.Stat(newGatewayDir)
+		moveFiles := err == nil && gd != nil && gd.IsDir() && nerr != nil && ngd == nil
+		if moveFiles {
+			// renaming to new place
+			err = os.Rename(gatewayDir, newGatewayDir)
+			if err != nil {
+				return err
+			}
+			// generate a router id and store it for subsequent template updates
+			routerId := newUUID()
+			err = ioutil.WriteFile(newGatewayDir+"/config/routerid.txt", []byte(routerId), 0644)
+			if err != nil {
+				return err
+			}
+			updateFileContent := func(fileName, oldPath, newPath string) error {
+				content, err := ioutil.ReadFile(fileName)
+				if err != nil {
+					return err
+				}
+				updatedContent := strings.ReplaceAll(string(content), oldPath, newPath)
+				err = ioutil.WriteFile(fileName, []byte(updatedContent), 0)
+				if err != nil {
+					return err
+				}
+				return nil
+			}
+			// Updating paths in service files
+			err = updateFileContent(fmt.Sprintf("%s/user/%s.service", newGatewayDir, gw.GatewayName), getDataHome()+"/skupper/", getDataHome()+gatewayClusterDir)
+			if err != nil {
+				return err
+			}
+			err = updateFileContent(getConfigHome()+"/systemd/user/"+gw.GatewayName+".service", getDataHome()+"/skupper/", getDataHome()+gatewayClusterDir)
+			if err != nil {
+				return err
+			}
+
+			cmd := exec.Command("systemctl", "--user", "daemon-reload")
+			err = cmd.Run()
+			if err != nil {
+				return fmt.Errorf("Unable to user service daemon-reload: %w", err)
+			}
+			cmd = exec.Command("systemctl", "--user", "restart", gw.GatewayName+".service")
+			err = cmd.Run()
+			if err != nil {
+				return fmt.Errorf("Unable to user service restart: %w", err)
+			}
+		}
+
+		// updating router config to fix bad template issues
+		configmap, err := kube.GetConfigMap(gatewayPrefix+gw.GatewayName, cli.GetNamespace(), cli.KubeClient)
+		if err != nil {
+			return err
+		}
+		gatewayConfig, err := qdr.GetRouterConfigFromConfigMap(configmap)
+		if err != nil {
+			return err
+		}
+		// updating version
+		sm := qdr.SiteMetadata{}
+		err = json.Unmarshal([]byte(gatewayConfig.Metadata.Metadata), &sm)
+		if err != nil {
+			return err
+		}
+		sm.Version = Version
+		smStr, err := json.Marshal(sm)
+		if err != nil {
+			return err
+		}
+		gatewayConfig.Metadata.Metadata = string(smStr)
+		// updating tcp listeners
+		newTcpListeners := qdr.TcpEndpointMap{}
+		for k, v := range gatewayConfig.Bridges.TcpListeners {
+			name := fmt.Sprintf("%s:%d", k, svcMap[v.Address].Ports[0])
+			v.Name = name
+			v.Address = fmt.Sprintf("%s:%d", v.Address, svcMap[v.Address].Ports[0])
+			newTcpListeners[name] = v
+		}
+		gatewayConfig.Bridges.TcpListeners = newTcpListeners
+		// updating tcp connectors
+		newTcpConnectors := qdr.TcpEndpointMap{}
+		for k, v := range gatewayConfig.Bridges.TcpConnectors {
+			name := fmt.Sprintf("%s:%d", k, svcMap[v.Address].Ports[0])
+			v.Name = name
+			v.Address = fmt.Sprintf("%s:%d", v.Address, svcMap[v.Address].Ports[0])
+			newTcpConnectors[name] = v
+		}
+		gatewayConfig.Bridges.TcpConnectors = newTcpConnectors
+		// updating http listeners
+		newHttpListeners := qdr.HttpEndpointMap{}
+		for k, v := range gatewayConfig.Bridges.HttpListeners {
+			name := fmt.Sprintf("%s:%d", k, svcMap[v.Address].Ports[0])
+			v.Name = name
+			v.Address = fmt.Sprintf("%s:%d", v.Address, svcMap[v.Address].Ports[0])
+			newHttpListeners[name] = v
+		}
+		gatewayConfig.Bridges.HttpListeners = newHttpListeners
+		// updating tcp connectors
+		newHttpConnectors := qdr.HttpEndpointMap{}
+		for k, v := range gatewayConfig.Bridges.HttpConnectors {
+			name := fmt.Sprintf("%s:%d", k, svcMap[v.Address].Ports[0])
+			v.Name = name
+			v.Address = fmt.Sprintf("%s:%d", v.Address, svcMap[v.Address].Ports[0])
+			newHttpConnectors[name] = v
+		}
+		gatewayConfig.Bridges.HttpConnectors = newHttpConnectors
+
+		// updating configmap
+		_ = gatewayConfig.WriteToConfigMap(configmap)
+		_, err = cli.KubeClient.CoreV1().ConfigMaps(cli.GetNamespace()).Update(configmap)
+		if err != nil {
+			return fmt.Errorf("Failed to update gateway config map: %s", err)
+		}
+		if err != nil {
+			return err
+		}
+
+		// if it is a local gateway
+		_, err = os.Stat(newGatewayDir + "/config/qdrouterd.json")
+		if err == nil {
+			// for update gatewayType would be "service"
+			err = updateLocalGatewayConfig(newGatewayDir, "service", *gatewayConfig)
+			if err != nil {
+				return err
+			}
+		}
+	}
+
+	return nil
+}
+
+func (cli *VanClient) restartRouter(namespace string) error {
+	router, err := cli.KubeClient.AppsV1().Deployments(namespace).Get(types.TransportDeploymentName, metav1.GetOptions{})
+	if err != nil {
+		return err
+	}
+	touch(router)
+	_, err = cli.KubeClient.AppsV1().Deployments(namespace).Update(router)
+	return err
+}
 
 func (cli *VanClient) RouterUpdateLogging(ctx context.Context, settings *corev1.ConfigMap, hup bool) (bool, error) {
 	siteConfig, err := cli.SiteConfigInspect(ctx, settings)
@@ -527,12 +790,7 @@ func (cli *VanClient) RouterUpdateLogging(ctx context.Context, settings *corev1.
 			return false, err
 		}
 		if hup {
-			router, err := cli.KubeClient.AppsV1().Deployments(settings.ObjectMeta.Namespace).Get(types.TransportDeploymentName, metav1.GetOptions{})
-			if err != nil {
-				return false, err
-			}
-			touch(router)
-			_, err = cli.KubeClient.AppsV1().Deployments(settings.ObjectMeta.Namespace).Update(router)
+			err = cli.restartRouter(settings.ObjectMeta.Namespace)
 			if err != nil {
 				return false, err
 			}
@@ -686,3 +944,102 @@ func (cli *VanClient) getTransportHosts(namespace string) ([]string, error) {
 func qualifiedServiceName(name string, namespace string) string {
 	return name + "." + namespace + ".svc.cluster.local"
 }
+
+func (cli *VanClient) addClaimsPortsToControllerService(ctx context.Context, namespace string) error {
+	svc, err := cli.KubeClient.CoreV1().Services(namespace).Get(types.ControllerServiceName, metav1.GetOptions{})
+	if err != nil {
+		return err
+	}
+	svc.Spec.Ports = append(svc.Spec.Ports, corev1.ServicePort{
+		Name:     types.ClaimRedemptionPortName,
+		Protocol: "TCP",
+		Port:     types.ClaimRedemptionPort,
+	})
+	_, err = cli.KubeClient.CoreV1().Services(namespace).Update(svc)
+	if err != nil {
+		return err
+	}
+	return nil
+}
+
+func (cli *VanClient) createClaimsServerSecret(ctx context.Context, namespace string, owner *metav1.OwnerReference, usingRoutes bool) error {
+	cred := types.Credential{
+		CA:          types.SiteCaSecret,
+		Name:        types.ClaimsServerSecret,
+		Subject:     types.ControllerServiceName,
+		Hosts:       []string{types.ControllerServiceName + "." + namespace},
+		ConnectJson: false,
+	}
+	if usingRoutes {
+		rte, err := kube.GetRoute(types.ClaimRedemptionRouteName, namespace, cli.RouteClient)
+		if err == nil {
+			cred.Hosts = append(cred.Hosts, rte.Spec.Host)
+		} else {
+			log.Printf("Failed to retrieve route %q: %s", types.ClaimRedemptionRouteName, err.Error())
+		}
+	} else {
+		err := cli.appendLoadBalancerHostOrIp(types.ControllerServiceName, namespace, &cred)
+		if err != nil {
+			return err
+		}
+	}
+	_, err := kube.NewSecret(cred, owner, namespace, cli.KubeClient)
+	if err != nil && !errors.IsAlreadyExists(err) {
+		return err
+	}
+	return nil
+}
+
+func (cli *VanClient) createClaimsRedemptionRoute(ctx context.Context, namespace string) error {
+	route := &routev1.Route{
+		TypeMeta: metav1.TypeMeta{
+			APIVersion: "v1",
+			Kind:       "Route",
+		},
+		ObjectMeta: metav1.ObjectMeta{
+			Name: types.ClaimRedemptionRouteName,
+		},
+		Spec: routev1.RouteSpec{
+			Path: "",
+			Port: &routev1.RoutePort{
+				TargetPort: intstr.FromString(types.ClaimRedemptionPortName),
+			},
+			To: routev1.RouteTargetReference{
+				Kind: "Service",
+				Name: types.ControllerServiceName,
+			},
+			TLS: &routev1.TLSConfig{
+				Termination:                   routev1.TLSTerminationPassthrough,
+				InsecureEdgeTerminationPolicy: routev1.InsecureEdgeTerminationPolicyRedirect,
+			},
+		},
+	}
+	_, err := kube.CreateRoute(route, namespace, cli.RouteClient)
+	if err != nil && !errors.IsAlreadyExists(err) {
+		return err
+	}
+	return nil
+}
+
+func (cli *VanClient) restartController(namespace string) error {
+	controller, err := cli.KubeClient.AppsV1().Deployments(namespace).Get(types.ControllerDeploymentName, metav1.GetOptions{})
+	if err != nil {
+		return err
+	}
+	touch(controller)
+	_, err = cli.KubeClient.AppsV1().Deployments(namespace).Update(controller)
+	return err
+}
+
+func (cli *VanClient) GetSiteMetadata() (*qdr.SiteMetadata, error) {
+	configmap, err := cli.KubeClient.CoreV1().ConfigMaps(cli.Namespace).Get(types.TransportConfigMapName, metav1.GetOptions{})
+	if err != nil {
+		return nil, err
+	}
+	config, err := qdr.GetRouterConfigFromConfigMap(configmap)
+	if err != nil {
+		return nil, err
+	}
+	metadata := config.GetSiteMetadata()
+	return &metadata, nil
+}
